@@ -1,16 +1,18 @@
 from angr.errors import SimEngineError, SimMemoryError
 from angr.analyses.bindiff import differing_constants
 from angr.analyses.bindiff import UnmatchedStatementsException
+from angr.misc.loggers import CuteFormatter
 from angr import Analysis, register_analysis
 
 from . import chain_builder
 from . import gadget_analyzer
 from . import common
+from .arch import get_arch
 
 import pickle
 import inspect
 import logging
-import progressbar
+import tqdm
 
 from .errors import RopException
 from .rop_gadget import RopGadget, StackPivot
@@ -18,16 +20,23 @@ from .rop_gadget import RopGadget, StackPivot
 from multiprocessing import Pool
 
 l = logging.getLogger('angrop.rop')
+logging.getLogger('pyvex.lifting').setLevel("ERROR")
 
 
 _global_gadget_analyzer = None
 
+# disable loggers in each worker
+def _disable_loggers():
+    for handler in logging.root.handlers:
+        if type(handler.formatter) == CuteFormatter:
+            logging.root.removeHandler(handler)
+            return
 
 # global initializer for multiprocessing
 def _set_global_gadget_analyzer(rop_gadget_analyzer):
     global _global_gadget_analyzer
     _global_gadget_analyzer = rop_gadget_analyzer
-
+    _disable_loggers()
 
 def run_worker(addr):
     return _global_gadget_analyzer.analyze_gadget(addr)
@@ -45,7 +54,7 @@ class ROP(Analysis):
     Additionally, all public methods from ChainBuilder are copied into ROP.
     """
 
-    def __init__(self, only_check_near_rets=True, max_block_size=20, max_sym_mem_accesses=4, fast_mode=None):
+    def __init__(self, only_check_near_rets=True, max_block_size=None, max_sym_mem_access=None, fast_mode=None, rebase=True, is_thumb=False):
         """
         Initializes the rop gadget finder
         :param only_check_near_rets: If true we skip blocks that are not near rets
@@ -53,29 +62,31 @@ class ROP(Analysis):
                                gadgets so we limit the size we consider
         :param fast_mode: if set to True sets options to run fast, if set to False sets options to find more gadgets
                           if set to None makes a decision based on the size of the binary
+        :param rebase:    if set to True, angrop will try to rebase the gadgets with its best effort
+                          if set to False, angrop will use the memory mapping in angr in the ropchain
+        :param is_thumb:  execute ROP chain in thumb mode. Only makes difference on ARM architecture.
+                          angrop does not switch mode within a rop chain
         :return:
         """
 
         # params
-        self._max_block_size = max_block_size
+        self.arch = get_arch(self.project)
         self._only_check_near_rets = only_check_near_rets
-        self._max_sym_mem_accesses = max_sym_mem_accesses
+        self._rebase = rebase
 
-        a = self.project.arch
-        self._sp_reg = a.register_names[a.sp_offset]
-        self._ip_reg = a.register_names[a.ip_offset]
-        self._base_pointer = a.register_names[a.bp_offset]
-
-        # get list of multipurpose registers
-        self._reg_list = a.default_symbolic_registers
-        # prune the register list of the instruction pointer and the stack pointer
-        self._reg_list = [r for r in self._reg_list if r not in (self._sp_reg, self._ip_reg)]
+        # override parameters
+        if max_block_size:
+            self.arch.max_block_size = max_block_size
+        if max_sym_mem_access:
+            self.arch.max_sym_mem_access = max_sym_mem_access
+        if is_thumb:
+            self.arch.is_thumb = is_thumb
 
         # get ret locations
-        self._ret_locations = self._get_ret_locations()
+        self._ret_locations = None
 
         # list of RopGadget's
-        self.gadgets = []
+        self._gadgets = []
         self.stack_pivots = []
         self._duplicates = []
 
@@ -83,28 +94,11 @@ class ROP(Analysis):
         self.badbytes = []
         self.roparg_filler = None
 
-        num_to_check = self._num_addresses_to_check()
-        # fast mode
-        if fast_mode is None:
-            if num_to_check > 20000:
-                fast_mode = True
-                l.warning("Enabling fast mode for large binary")
-            else:
-                fast_mode = False
         self._fast_mode = fast_mode
 
-        if self._fast_mode:
-            self._max_block_size = 12
-            self._max_sym_mem_accesses = 1
-            # Recalculate num addresses to check based on fast_mode settings
-            num_to_check = self._num_addresses_to_check()
-
-        l.info("There are %d addresses within %d bytes of a ret",
-               num_to_check, self._max_block_size)
-
         # gadget analyzer
-        self._gadget_analyzer = gadget_analyzer.GadgetAnalyzer(self.project, self._reg_list, self._max_block_size,
-                                                               self._fast_mode, self._max_sym_mem_accesses)
+        self._gadget_analyzer = None
+
         # chain builder
         self._chain_builder = None
 
@@ -116,14 +110,43 @@ class ROP(Analysis):
         logging.getLogger('pyvex.lifting.libvex').setLevel(logging.CRITICAL)
         logging.getLogger('angr.procedures.cgc.deallocate').setLevel(logging.CRITICAL)
 
+    @property
+    def gadgets(self):
+        return [x for x in self._gadgets if not self._contain_badbytes(x.addr)]
+
+    def _initialize_gadget_analyzer(self):
+
+        # find locations to analyze
+        self._ret_locations = self._get_ret_locations()
+        num_to_check = self._num_addresses_to_check()
+
+        # fast mode
+        if self._fast_mode is None:
+            if num_to_check > 20000:
+                self._fast_mode = True
+                l.warning("Enabling fast mode for large binary")
+            else:
+                self._fast_mode = False
+        if self._fast_mode:
+            self.arch.max_block_size = 12
+            self.arch.max_sym_mem_access = 1
+            # Recalculate num addresses to check based on fast_mode settings
+            num_to_check = self._num_addresses_to_check()
+
+        l.info("There are %d addresses within %d bytes of a ret",
+               num_to_check, self.arch.max_block_size)
+
+        self._gadget_analyzer = gadget_analyzer.GadgetAnalyzer(self.project, self._fast_mode, arch=self.arch)
+
     def find_gadgets(self, processes=4, show_progress=True):
         """
         Finds all the gadgets in the binary by calling analyze_gadget on every address near a ret.
-        Saves gadgets in self.gadgets
+        Saves gadgets in self._gadgets
         Saves stack pivots in self.stack_pivots
         :param processes: number of processes to use
         """
-        self.gadgets = []
+        self._initialize_gadget_analyzer()
+        self._gadgets = []
 
         pool = Pool(processes=processes, initializer=_set_global_gadget_analyzer, initargs=(self._gadget_analyzer,))
 
@@ -131,23 +154,23 @@ class ROP(Analysis):
         for gadget in it:
             if gadget is not None:
                 if isinstance(gadget, RopGadget):
-                    self.gadgets.append(gadget)
+                    self._gadgets.append(gadget)
                 elif isinstance(gadget, StackPivot):
                     self.stack_pivots.append(gadget)
 
         pool.close()
 
         # fix up gadgets from cache
-        for g in self.gadgets:
+        for g in self._gadgets:
             if g.addr in self._cache:
                 dups = {g.addr}
                 for addr in self._cache[g.addr]:
                     dups.add(addr)
                     g_copy = g.copy()
                     g_copy.addr = addr
-                    self.gadgets.append(g_copy)
+                    self._gadgets.append(g_copy)
                 self._duplicates.append(dups)
-        self.gadgets = sorted(self.gadgets, key=lambda x: x.addr)
+        self._gadgets = sorted(self._gadgets, key=lambda x: x.addr)
         self._reload_chain_funcs()
 
     def find_gadgets_single_threaded(self, show_progress=True):
@@ -156,28 +179,29 @@ class ROP(Analysis):
         Saves gadgets in self.gadgets
         Saves stack pivots in self.stack_pivots
         """
-        self.gadgets = []
+        self._initialize_gadget_analyzer()
+        self._gadgets = []
 
         _set_global_gadget_analyzer(self._gadget_analyzer)
         for _, addr in enumerate(self._addresses_to_check_with_caching(show_progress)):
             gadget = _global_gadget_analyzer.analyze_gadget(addr)
             if gadget is not None:
                 if isinstance(gadget, RopGadget):
-                    self.gadgets.append(gadget)
+                    self._gadgets.append(gadget)
                 elif isinstance(gadget, StackPivot):
                     self.stack_pivots.append(gadget)
 
         # fix up gadgets from cache
-        for g in self.gadgets:
+        for g in self._gadgets:
             if g.addr in self._cache:
                 dups = {g.addr}
                 for addr in self._cache[g.addr]:
                     dups.add(addr)
                     g_copy = g.copy()
                     g_copy.addr = addr
-                    self.gadgets.append(g_copy)
+                    self._gadgets.append(g_copy)
                 self._duplicates.append(dups)
-        self.gadgets = sorted(self.gadgets, key=lambda x: x.addr)
+        self._gadgets = sorted(self._gadgets, key=lambda x: x.addr)
         self._reload_chain_funcs()
 
     def save_gadgets(self, path):
@@ -204,8 +228,9 @@ class ROP(Analysis):
         if not isinstance(badbytes, list):
             print("Require a list, e.g: [0x00, 0x09]")
             return
+        badbytes = [x if type(x) == int else ord(x) for x in badbytes]
         self.badbytes = badbytes
-        if len(self.gadgets) > 0:
+        if len(self._gadgets) > 0:
             self.chain_builder._set_badbytes(self.badbytes)
 
     def set_roparg_filler(self, roparg_filler):
@@ -221,7 +246,7 @@ class ROP(Analysis):
             return
 
         self.roparg_filler = roparg_filler
-        if len(self.gadgets) > 0:
+        if len(self._gadgets) > 0:
             self.chain_builder._set_roparg_filler(self.roparg_filler)
 
     def get_badbytes(self):
@@ -232,10 +257,12 @@ class ROP(Analysis):
         return self.badbytes
 
     def _get_cache_tuple(self):
-        return self.gadgets, self.stack_pivots, self._duplicates
+        return self._gadgets, self.stack_pivots, self._duplicates, self._ret_locations, self._fast_mode, \
+                self.arch, self._gadget_analyzer
 
     def _load_cache_tuple(self, cache_tuple):
-        self.gadgets, self.stack_pivots, self._duplicates = cache_tuple
+        self._gadgets, self.stack_pivots, self._duplicates, self._ret_locations, self._fast_mode, \
+        self.arch, self._gadget_analyzer = cache_tuple
         self._reload_chain_funcs()
 
     def _reload_chain_funcs(self):
@@ -248,13 +275,12 @@ class ROP(Analysis):
     def chain_builder(self):
         if self._chain_builder is not None:
             return self._chain_builder
-        elif len(self.gadgets) > 0:
-            self._chain_builder = chain_builder.ChainBuilder(self.project, self.gadgets, self._duplicates,
-                                                             self._reg_list, self._base_pointer, self.badbytes,
-                                                             self.roparg_filler)
-            return self._chain_builder
-        else:
-            raise Exception("No gadgets available, call find_gadgets() or load_gadgets() if you haven't already.")
+        if len(self._gadgets) == 0:
+            l.warning("Could not find gadgets for %s, check your badbytes and make sure find_gadgets() or load_gadgets() was called.", self.project)
+        self._chain_builder = chain_builder.ChainBuilder(self.project, self.gadgets, self._duplicates,
+                                                         self.arch.reg_list, self.arch.base_pointer, self.badbytes,
+                                                         self.roparg_filler, rebase=self._rebase)
+        return self._chain_builder
 
     def _block_has_ip_relative(self, addr, bl):
         """
@@ -279,20 +305,18 @@ class ROP(Analysis):
 
     def _addresses_to_check_with_caching(self, show_progress=True):
         num_addrs = self._num_addresses_to_check()
-        widgets = ['ROP: ', progressbar.Percentage(), ' ',
-                   progressbar.Bar(marker=progressbar.RotatingMarker()),
-                   ' ', progressbar.ETA(), ' ', progressbar.FileTransferSpeed()]
-        progress = progressbar.ProgressBar(widgets=widgets, maxval=num_addrs)
+        self._cache = {}
+        seen = {}
+
+        iterable = self._addresses_to_check()
         if show_progress:
-            progress.start()
-        self._cache = dict()
-        seen = dict()
-        for i, a in enumerate(self._addresses_to_check()):
-            if show_progress:
-                progress.update(i)
+            iterable = tqdm.tqdm(iterable=iterable, smoothing=0, total=num_addrs,
+                                 desc="ROP", maxinterval=0.5, dynamic_ncols=True)
+
+        for a in iterable:
             try:
                 bl = self.project.factory.block(a)
-                if bl.size > self._max_block_size:
+                if bl.size > self.arch.max_block_size:
                     continue
                 block_data = bl.bytes
             except (SimEngineError, SimMemoryError):
@@ -307,8 +331,6 @@ class ROP(Analysis):
                     seen[block_data] = a
                     self._cache[a] = set()
                 yield a
-        if show_progress:
-            progress.finish()
 
     def _addresses_to_check(self):
         """
@@ -316,8 +338,8 @@ class ROP(Analysis):
         """
         if self._only_check_near_rets:
             # align block size
-            alignment = self.project.arch.instruction_alignment
-            block_size = (self._max_block_size & ((1 << self.project.arch.bits) - alignment)) + alignment
+            alignment = self.arch.alignment
+            block_size = (self.arch.max_block_size & ((1 << self.project.arch.bits) - alignment)) + alignment
             slices = [(addr-block_size, addr) for addr in self._ret_locations]
             current_addr = 0
             for st, _ in slices:
@@ -347,7 +369,7 @@ class ROP(Analysis):
                 if segment.is_executable:
                     num += (segment.max_addr - segment.min_addr)
             return num
-                        
+
     def _get_ret_locations(self):
         """
         :return: all the locations in the binary with a ret instruction
@@ -364,10 +386,7 @@ class ROP(Analysis):
             if segment.is_executable:
                 num_bytes = segment.max_addr-segment.min_addr
 
-                alignment = self.project.arch.instruction_alignment
-                # hack for arm thumb
-                if self.project.arch.linux_name == "aarch64" or self.project.arch.linux_name == "arm":
-                    alignment = 1
+                alignment = self.arch.alignment
 
                 # iterate through the code looking for rets
                 for addr in range(segment.min_addr, segment.min_addr + num_bytes, alignment):
@@ -432,5 +451,16 @@ class ROP(Analysis):
             return True
         return False
 
+    # inspired by ropper
+    def _contain_badbytes(self, addr):
+        n_bytes = self.project.arch.bytes
+
+        for b in self.badbytes:
+            tmp_addr = addr
+            for _ in range(n_bytes):
+                if (tmp_addr & 0xff) == b:
+                    return True
+                tmp_addr >>= 8
+        return False
 
 register_analysis(ROP, 'ROP')
